"""
Plotting module for PyTorch prototyping

:author: Julian M. Kleber
"""
from typing import Type, Optional, Dict, Any, List

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

from collections import ChainMap


def plot_range_band(result: List[Dict[str, float]], key_val: str) -> None:

    arr_res = np.zeros((len(result), len(result[0][key_val])))
    print(arr_res.shape)
    for i, res in enumerate(result):
        print((len(res[key_val])))
        arr_res[i, :] = res[key_val] #TOOD fix the shapes


def get_stacked_list(
    path_to_directory: str, column_name: str, num_cv: int, json_name: str
) -> List[Dict[str, float]]:

    results = []

    for i in range(num_cv):
        try:
            result = parse_acc_list_json(
                path_to_json=path_to_directory + f"CV_{i}/" + json_name
            )
            results.append(result)

        except Exception as exc:
            print(exc)
            continue

    return results


def parse_old_file_format_for_plot(stacked_list: List[float]) -> List[List[float]]:
    """
    The parse_old_file_format_for_plot function takes in a path to a json file and returns the following:

    :param path_to_json:str: Used to Specify the path to the json file that we want to parse.
    :return: A list of dictionaries.

    :doc-author: Julian M. Kleber
    """

    result_acc = parse_acc_list_json(path_to_json=path_to_json)
    train_frames_acc = parse_min_max_avg(result_acc["Train_acc"])
    test_frames_acc = parse_min_max_avg(result_acc["Test_acc"])
    return train_frames_acc, test_frames_acc


def parse_acc_list_json(path_to_json: str) -> Dict[str, Any]:
    """
    The parse_acc_list_json function takes in a path to a json file and returns the contents of that json file as a dictionary.
    The function also parses the "Acc" key in the dictionary, which contains lists of tuples containing train and test accuracy values.
    The function then separates these tuples into two separate lists, one for train accuracy values and one for test accuracy values.
    These new lists are added to the original dictionary under keys "Acc_train" and "Acc_val", respectively.

    :param path_to_json:str: Used to Specify the path to the json file.
    :return: A dictionary with the following keys:.

    :doc-author: Trelent
    """

    result = load_json_from_file(path_to_json)
    acc_store = result["Acc"]

    train_acc = []
    test_acc = []

    for i in range(len(acc_store)):
        train, test = acc_store[i]
        train_acc.append(train)
        test_acc.append(test)

    del result["Acc"]
    print(f"Len of train, {len(train_acc)}")
    result["Acc_train"] = train_acc
    result["Acc_val"] = test_acc

    return result


def parse_min_max_avg(result_list: List[List[float]]) -> List[float]:
    """
    The parse_min function takes a list of lists and returns the minimum value for each sublist.

    :param result_list:List[List[float]]: Used to Store the results of the simulation.
    :return: A list of the minimum values for each step.

    :doc-author: Trelent
    """

    minima = []
    maxima = []
    averages = []
    for i in range(len(result_list[0])):

        step_list = result_list[:, i]
        minimum = get_min(step_list=step_list)
        maximum = get_max(step_list=step_list)
        average = get_avg(step_list=step_list)
        minima.append(minimum)
        maxima.append(maximum)
        averages.append(average)
    return [minima, maxima, averages]


def get_avg(step_list: List[float]) -> float:
    """
    The get_avg function takes a list of floats and returns the average value.


    :param step_list:List[float]: Used to Tell the function that it will be taking a list of floats as an argument.
    :return: The mean of the step_list.

    :doc-author: Julian M. Kleber
    """

    return np.mean(step_list)


def get_max(step_list: List[float]) -> float:
    """
    The get_max function takes a list of floats and returns the maximum value in that list.


    :param step_list:List[float]: Used to Tell the function that step_list is a list of floats.
    :return: The maximum value in the list.

    :doc-author: Trelent
    """

    return np.max(step_list)


def get_min(step_list: List[float]) -> float:
    """
    The get_min function takes a list of floats and returns the minimum value in that list.

    :param step_list:List[float]: Used to Specify the type of parameter that is being passed into the function.
    :return: The minimum value in the step_list.

    :doc-author: Trelent
    """

    return np.min(step_list)


if __name__ == "__main__":
    from amarium.utils import load_json_from_file
    import pandas as pd

    path_to_directory = (
        "/home/dev/carate/carate/carate_config_files/Classification/PROTEINS_10/data/"
    )
    result = get_stacked_list(
        path_to_directory=path_to_directory,
        column_name="Acc",
        num_cv=5,
        json_name="PROTEINS_Epoch_4980.json",
    )
    plot_range_band(result, key_val="Acc_train")



import torch
from abc import ABC, abstractmethod


class Model(torch.nn.Module):
    @abstractmethod
    def __init__(self, dim: int, num_classes: int, num_features: int) -> None:
        super(Model, self).__init__()
        self.num_classes = num_classes
        self.num_features = num_features
        self.dim = dim

    @abstractmethod
    def forward(
        self, x: int, edge_index: int, batch: int, edge_weight=None
    ) -> torch.Tensor:
        pass  # pragma: no cover



"""
Module for handling config files used to start 
the training and evaluation process 

:author: Julian M. Kleber
"""

from typing import Optional, List, Dict, Any

import subprocess
from amarium.utils import save_json_to_file


def convert_py_to_json(
    file_name: str, out_name: Optional[str] = None
) -> Dict[str, Any]:
    """
    The convert_py_to_json function takes in a file name and an output file name.
    It reads the input file, which is assumed to be a .py config file with key value pairs separated by spaces.
    The function then converts the key value pairs into a dictionary and saves it as json format in the output filename.

    :param file_name:str: Used to Specify the file name of the.
    :param out_name:str: Used to Specify the name of the output file.
    :return: A dictionary with the same information as the file.

    :doc-author: Julian M. Kleber
    """

    json_dict: Dict[str, Any] = {}
    lines = read_file(file_name)

    for line in lines:
        tmp = line.split(" ")
        if tmp[2] == "None" or tmp[2] == "None":
            json_dict[tmp[0]] = None
        else:
            json_dict[tmp[0]] = tmp[2]

    if out_name is not None:
        save_json_to_file(json_dict, file_name=out_name)

    return json_dict


def read_file(file_name: str) -> List[str]:
    """
    The read_file function takes a file name as input and returns the contents of that file as a list of strings.
    The function also formats the code using black before reading it in.

    :param file_name:str: Used to Specify the file name.
    :return: A list of strings.

    :doc-author: Julian M. Kleber
    """

    subprocess.run(
        ["black", file_name], capture_output=True
    )  # format file to avoid misunderstandings
    with open(file_name, "r", encoding="utf-8") as file:
        raw = file.readlines()
    result = sanitize_raw_py(raw)
    return result


def sanitize_raw_py(raw_input: List[str]) -> List[str]:
    """
    The sanitize_raw_py function takes in a list of strings and returns a list of strings.
    The function removes all newline characters from the input, as well as any quotation marks or apostrophes.

    :param raw_input:list[str]: Used to Specify the type of data that is expected to be passed into the function.
    :return: A list of strings.

    :doc-author: Julian M. Kleber
    """

    result = []
    for line in raw_input:
        if line == "\n":
            continue
        else:
            result.append(line.replace("\n", "").replace(
                '"', "").replace("'", ""))
    return result



dataset_name = "Yeast"
num_classes = 2
num_features = 46
model = "cgc_classification"
evaluation = "classification"
optimizer = "adams"  # defaults to adams optimizer
net_dimension = 364
learning_rate = 0.0005
dataset_save_path = "t./data"
test_ratio = 20
batch_size = 64
shuffle = True
gamma = 51
num_epoch = 300
num_cv = 5
result_save_dir = "./Yeast"
data_loader = "StandardTUD"
model_save_freq = 1



"""Module providing a CLI 
interface for training models#

:author: Julian  M. Kleber
"""
from typing import Type
import click

from carate.run import RunInitializer, Run


@click.command()
@click.option("-c", help="Path to config file")
def train_algorithm(c: str) -> None:
    """
    The train_algorithm function takes in a config file and an output directory.
    It then runs the algorithm using the configuration specified in the config file,
    and saves all of its outputs to the output directory.

    :param c:str: Used to Specify the path to the configuration file.
    :param o:str: Used to Specify the output directory.
    :return: None.

    :doc-author: Julian M. Kleber
    """
    if c is None:
        raise RuntimeError("Please provide the path to a config file.")
    config_filepath = c
    runner = RunInitializer.from_file(config_filepath=config_filepath)

    runner.run()


if __name__ == "__main__":
    train_algorithm()



"""
Evaulation object for classification
"""
from typing import Type

import torch
from carate.evaluation.base import Evaluation
from carate.load_data import DatasetObject


class ClassificationEvaluation(Evaluation):
    def __init__(
        self,
        dataset_name: str,
        dataset_save_path: str,
        result_save_dir: str,
        model_net: torch.nn.Module,
        optimizer: torch.optim.Optimizer,
        data_set: DatasetObject,
        model_save_freq: int,
        test_ratio: int,
        num_epoch: int = 150,
        num_cv: int = 5,
        num_classes: int = 2,
        out_dir: str = r"./out",
        batch_size: int = 64,
        shuffle: bool = True,
        override: bool = True,
    ):
        """

        :param self: Used to Refer to the object instance itself, and
        is used to access variables that belongs to the class.
        :param model: Used to Specify the model that will be trained.
        :param optimizer: Used to Define the optimizer that will be used
        to train the model.
        :param data_set:Type[DatasetObject]: Used to Specify the type of data
        loader that is used. Is implemented according to
        the interface given in load_data.py by the class DatasetObject.load_data().

        :param epoch:int=150: Used to Set the number of epochs to train for.
        :param num_cv:int=5: Used to Specify the number of cross validations that will be used in the training process.
        :param num_classes:int=2: Used to Define the number of classes in the dataset.
        :param out_dir:str="out": Used to Specify the directory where the output of your training will be stored.
        :return: The following:.

        :doc-author: Julian M. Kleber
        """
        self.dataset_name = dataset_name
        self.dataset_save_path = dataset_save_path
        self.test_ratio = test_ratio
        self.num_epoch = num_epoch
        self.model_net = model_net
        self.optimizer = optimizer
        self.num_classes = num_classes
        self.num_cv = num_cv
        self.out_dir = out_dir
        self.data_set = data_set
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.result_save_dir = result_save_dir
        self.model_save_freq = model_save_freq
        self.override = override

    def __repr__(self) -> str:
        return "Classification Evaluation Object"



"""
Evaulation object for classification
"""
import torch
import numpy as np
import numpy.typing as npt
from typing import Type, Any, Tuple, Dict

from carate.evaluation.base import Evaluation
from carate.load_data import DatasetObject
from carate.utils.model_files import save_model_parameters
from carate.models.base_model import Model

# TODO Logging done right
import logging

logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class RegressionEvaluation(Evaluation):
    """
    Module that implements the Regression evaluation
    """

    def __init__(
        self,
        dataset_name: str,
        dataset_save_path: str,
        result_save_dir: str,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        data_set: DatasetObject,
        test_ratio: int,
        num_epoch: int = 150,
        num_cv: int = 5,
        num_classes: int = 2,
        out_dir: str = r"./out",
        batch_size: int = 64,
        shuffle: bool = True,
        model_save_freq: int = 100,
        override: bool = True,
    ):
        """
        The __init__ function is called when the class is instantiated.
        It sets up all the parameters needed for training and testing.

        :param self: Used to Refer to the current object.
        :param dataset_name:str: Used to Name the dataset.
        :param dataset_save_path:str: Used to Save the dataset object
         to a file.
        :param result_save_dir:str: Used to Save the results of the
        cross validation.
        :param model_net:Model: Used to Specify the model architecture.
        :param optimizer:torch.optim.Optimizer: Used to Define the optimizer
        used for training.
        :param data_set:DatasetObject: Used to Pass the dataset object
        to the class.
        :param test_ratio:int: Used to Determine the ratio of test data to
        training data.
        :param num_epoch:int=150: Used to Set the number of epochs to train for.
        :param num_cv:int=5: Used to Set the number of cross-validation folds.
        :param num_classes:int=2: Used to Set the number of classes in the dataset.
        :param out_dir:str=r"./out": Used to Define the directory where
        all results are saved.
        :param batch_size:int=64: Used to Set the batch size of the training data.
        :param shuffle:bool=True: Used to Shuffle the data set before splitting
        it into training and test sets.
        :param model_save_freq:int=100: Used to Save the model every 100 epochs.
        :param override:bool=True: Used to Override the results if they already exist.
        :param : Used to Set the number of epochs for training.
        :return: The object itself, which is then assigned to the variable "model_trainer".

        :doc-author: Trelent
        """
        """


        :doc-author: Julian M. Kleber
        """
        self.dataset_name = dataset_name
        self.dataset_save_path = dataset_save_path
        self.test_ratio = test_ratio
        self.num_epoch = num_epoch
        self.model_net = model_net
        self.optimizer = optimizer
        self.num_classes = num_classes
        self.num_cv = num_cv
        self.out_dir = out_dir
        self.data_set = data_set
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.result_save_dir = result_save_dir
        self.model_save_freq = model_save_freq
        self.override = override

    def cv(
        self,
        num_cv: int,
        num_epoch: int,
        num_classes: int,
        dataset_name: str,
        dataset_save_path: str,
        test_ratio: int,
        data_set: DatasetObject,
        shuffle: bool,
        batch_size: int,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        result_save_dir: str,
        model_save_freq: int,
        override: bool = True,
    ) -> Dict[str, Any]:

        # initialize
        (
            num_cv,
            num_epoch,
            num_classes,
            dataset_name,
            dataset_save_path,
            test_ratio,
            data_set,
            shuffle,
            batch_size,
            model_net,
            optimizer,
            device,
            result_save_dir,
            model_save_freq,
            override,
        ) = self._get_defaults(locals())

        # data container
        result = {}
        test_mse = []
        train_mae = []
        train_mse = []
        tmp = {}

        save_model_parameters(model_net=model_net, save_dir=result_save_dir)
        for i in range(num_cv):
            loaded_dataset: torch.utils.data.Dataset
            (
                train_loader,
                test_loader,
                loaded_data_set,
                train_dataset,
                test_dataset,
            ) = data_set.load_data(
                dataset_name=dataset_name,
                dataset_save_path=dataset_save_path,
                test_ratio=test_ratio,
                batch_size=batch_size,
                shuffle=shuffle,
            )
            del train_dataset, test_dataset
            norm_factor = self.__normalization_factor(
                data_set=loaded_data_set, num_classes=num_classes
            )
            for epoch in range(1, num_epoch + 1):
                train_mae_loss = self.train(
                    model_net=model_net,
                    epoch=epoch,
                    optimizer=optimizer,
                    train_loader=train_loader,
                    device=device,
                    norm_factor=norm_factor,
                    num_classes=num_classes,
                )
                train_mae_val, train_mse_val = self.test(
                    model_net=model_net,
                    test_loader=train_loader,
                    device=device,
                    norm_factor=norm_factor,
                    epoch=epoch,
                )
                test_mae_val, test_mse_val = self.test(
                    model_net=model_net,
                    test_loader=test_loader,
                    device=device,
                    norm_factor=norm_factor,
                    epoch=epoch,
                )
                train_mae.append(train_mae_val)
                train_mse.append(train_mse_val)
                test_mse.append(test_mae_val)
                test_mse.append(test_mse_val)
                logging.info(
                    "Epoch: {:03d}, Train MAE, MSE at epoch: ({:.7f}, {:.7f}), "
                    "Test MAE, MSE at epoch: ({:.7f}, {:.7f})".format(
                        epoch, train_mae_val, train_mse_val, test_mae_val, test_mse_val
                    )
                )
                torch.cuda.empty_cache()

                tmp["MAE Train"] = list(train_mae)
                tmp["MSE Train"] = list(train_mse)
                tmp["MAE Test"] = list(train_mae)
                tmp["MSE Test"] = list(train_mse)

                if epoch % model_save_freq == 0:
                    self.save_whole_checkpoint(
                        result_save_dir=result_save_dir,
                        dataset_name=dataset_name,
                        num_cv=i,
                        num_epoch=epoch,
                        model_net=model_net,
                        data=tmp,
                        optimizer=optimizer,
                        loss=train_mae_loss,
                        override=override,
                    )

            result[str(i)] = tmp
        return result

    def train(
        self,
        epoch: int,
        model_net: Model,
        device: torch.device,
        train_loader: torch.utils.data.DataLoader,
        optimizer: torch.optim.Optimizer,
        num_classes: int,
        **kwargs: Any,
    ) -> float:
        norm_factor = float(kwargs["norm_factor"])
        model_net.train()
        mse = 0
        for data in train_loader:
            data.x = data.x.type(torch.FloatTensor)
            data.y = (data.y.numpy()) / norm_factor
            data.y = torch.from_numpy(data.y).type(torch.FloatTensor)
            data.y = torch.nan_to_num(data.y.type(torch.FloatTensor))
            data = data.to(device)
            optimizer.zero_grad()
            output_probs = model_net(data.x, data.edge_index, data.batch)
            loss = torch.nn.MSELoss()
            if len(data.y.size()) == 1:
                loss = loss(output_probs, data.y[:])
            else:
                loss = loss(output_probs, data.y[:, :])
            mse += loss.item()
            loss.backward()
            optimizer.step()
            torch.cuda.empty_cache()
        return mse / len(train_loader)

    def test(
        self,
        test_loader: Type[torch.utils.data.DataLoader],
        epoch: int,
        model_net: Model,
        device: torch.device,
        **kwargs: Any,
    ) -> Tuple[float, float]:
        norm_factor = float(kwargs["norm_factor"])
        model_net.eval()
        mae = 0
        mse = 0
        for data in test_loader:
            data.x = data.x.type(torch.FloatTensor)
            data.y = data.y / norm_factor
            data.y = torch.nan_to_num(data.y.type(torch.FloatTensor))
            data = data.to(device)
            output_probs = model_net(data.x, data.edge_index, data.batch)
            loss_mae = torch.nn.L1Loss()
            mae += loss_mae(output_probs, data.y).item()
            loss = torch.nn.MSELoss()
            mse += loss_mae(output_probs, data.y).item()
            torch.cuda.empty_cache()
        return mae / len(test_loader), mse / len(
            test_loader
        )  # TODO verify if necessary

    def __normalization_factor(
        self, data_set: Any, num_classes: int
    ) -> npt.NDArray[np.float64]:

        y = np.zeros((len(data_set), 1, num_classes))
        for i in range(len(data_set)):
            y[i, :, :] = data_set[i].y
        norm_factor = np.zeros((num_classes))
        for i in range(num_classes):
            norm = np.linalg.norm(y[:, 0, i], ord=2)
            norm_factor[i] = norm
        return norm_factor

    def __repr__(self) -> str:
        return "Regression Evaluation Object"



"""
Module to handle optimizer initalization


@author Julian M. Kleber
"""
import torch
from typing import Type

from carate.models.base_model import Model


def get_optimizer(
    optimizer_str: str, model_net: Model, learning_rate: float
) -> torch.optim.Optimizer:
    """
    The get_optimizer function takes in a string and returns the corresponding optimizer.
        Args:
            optimizer_str (str): The name of the desired optimizer.

        Returns:
            Type[torch.optim.Optimizer]: The corresponding torch optimization function for the given string.

    :param optimizer_str:str: Used to Specify the type of optimizer we want to use.
    :param model: Used to Get the parameters of the model.
    :return: The optimizer of the network.

    :doc-author: Trelent
    """

    if optimizer_str == "adams":
        return torch.optim.Adam(model_net.parameters(), lr=learning_rate)



"""Default Interface defining the Base object for all other objects. 
Especially the __get_default_method is provided for all base objects. 

:author: Julian M. Kleber
"""
from typing import Type, Optional, List, Dict, Any

import logging


logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class DefaultObject:
    """
    The default object provides methods every class needs on top of the standard
    Python functionality.
    """

    def _get_defaults(self, method_arguments: Dict[Any, Any]) -> List[Any]:
        """
        The _get_defaults function takes a dictionary of arguments and returns a list of values.
        The function checks if the value is None, if it is none then it checks to see if that key exists in the instance variables.
        If so, then it will return the instance variable associated with that key. If not, then nothing happens and None gets returned.

        :param method_arguments:dict: Used to pass in the arguments that are passed into the method.
        :return: A list of values that are either none or the value provided in the method_arguments dictionary.

        :doc-author: Trelent
        """

        result = []
        instance_variables = vars(self)  # dictionary with instance variables
        instance_attributes = instance_variables.keys()

        for key, value in method_arguments.items():
            if key == "self":
                continue
            print(value, instance_variables[key])
            if value is None and key in instance_attributes:
                result.append(instance_variables[key])
            elif value != None:
                result.append(value)
            elif value is None and key not in instance_attributes:
                result.append(None)
            else:
                result.append(None)
        return result



dataset_name = "PROTEINS"
num_classes = 2
num_features = 3
model = "cgc_classification"
evaluation = "classification"
optimizer = "adams"  # defaults to adams optimizer
net_dimension = 364
learning_rate = 0.0005
dataset_save_path = "./data"
test_ratio = 20
batch_size = 64
shuffle = True
num_epoch = 5000
num_cv = 3
result_save_dir = "./PROTEINS_20"
data_set = "StandardTUD"
model_save_freq = 30

if __name__ == "__main__":

    from carate.run import RunInitializer

    config_filepath = "./PROTEINS.py"
    runner = RunInitializer.from_file(config_filepath=config_filepath)
    runner.run()



"""
Module for serialization and deserialization of inputs. The aim is to
keep web-first attitude, even though when using files locally. If there
is text files then there is a need to convert them.

@author = Julian M. Kleber
"""
import torch
from typing import Type, Optional, Dict, TypeVar, Any, Generic

from amarium.utils import convert_str_to_bool

from carate.evaluation import base, classification, regression
from carate.models import cgc_classification, cgc_regression
from carate.load_data import (
    DatasetObject,
    StandardPytorchGeometricDataset,
    StandardDatasetTUDataset,
    StandardDatasetMoleculeNet,
)
from carate.utils.convert_to_json import convert_py_to_json


EvaluationMap: Dict[str, base.Evaluation]
EVALUATION_MAP = {
    "regression": regression.RegressionEvaluation,
    "classification": classification.ClassificationEvaluation,
    "evaluation": base.Evaluation,
}

ModelMap: Dict[str, Any]
MODEL_MAP = {"cgc_classification": cgc_classification,
             "cgc_regression": cgc_regression}

DATA_SET_MAP: Dict[
    str,
    Type[StandardDatasetMoleculeNet]
    | Type[StandardPytorchGeometricDataset]
    | Type[StandardPytorchGeometricDataset],
]
DATA_SET_MAP = {
    "StandardPyG": StandardPytorchGeometricDataset,
    "StandardTUD": StandardDatasetTUDataset,
    "StandardMolNet": StandardDatasetMoleculeNet,
}


class Config:
    """
    The Config class is an object representation of the configuration of the model. It aims to provide a middle layer between
    some user input and the run interface. It is also possible to use it via the web because of the method overload of the constructor.

    :author: Julian M. Kleber
    """

    def __init__(
        self,
        dataset_name: str,
        num_features: int,
        num_classes: int,
        result_save_dir: str,
        model_save_freq: int,
        Evaluation: base.Evaluation,
        data_set: DatasetObject,
        model: Any,
        optimizer: str,
        net_dimension: int = 364,
        learning_rate: float = 0.0005,
        dataset_save_path: str = ".",
        test_ratio: int = 20,
        batch_size: int = 64,
        shuffle: bool = True,
        num_cv: int = 5,
        num_epoch: int = 150,
        override: bool = True,
    ):

        # fill with maps
        self.model = model
        self.optimizer = optimizer
        self.Evaluation = Evaluation
        self.data_set = data_set

        # model parameters
        self.dataset_name = dataset_name
        self.num_classes = num_classes
        self.num_features = num_features
        self.net_dimension = net_dimension
        self.learning_rate = learning_rate

        # evaluation parameters
        self.dataset_name = dataset_name
        self.dataset_save_path = dataset_save_path
        self.test_ratio = test_ratio
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.num_cv = num_cv
        self.num_epoch = num_epoch
        self.result_save_dir = result_save_dir
        self.model_save_freq = model_save_freq
        self.override = override


class ConfigInitializer:
    @classmethod
    def from_file(cls, file_name: str) -> Config:
        """
        The from_file function takes a file name as an argument and returns a Config object.
        The function reads the file, converts it to JSON, then uses the from_json method to create
        the Config object.

        :param cls: Used to create a new instance of the class.
        :param file_name:str: Used to specify the name of the file to be used.
        :return: A config object.

        :doc-author: Julian M. Kleber
        """

        json_object = convert_py_to_json(file_name)
        config_object = ConfigInitializer.from_json(json_object)
        return config_object

    @classmethod
    def from_json(cls, json_object: Dict[Any, Any]) -> Config:
        """
        The from_json function is a class method that takes in a json object and returns an instance of the Config class.
        The function is used to load the configuration from a file, which can be done by calling:
            config = Config.from_json(json_object)

        :param cls: Used to Create an instance of the class that is calling this method.
        :param json_object:dict: Used to Pass in the json object that is read from the file.
        :return: A class object.

        :doc-author: Julian M. Kleber
        """

        data_set = DATA_SET_MAP[json_object["data_set"]](
            dataset_save_path=json_object["dataset_save_path"],
            dataset_name=json_object["dataset_name"],
            test_ratio=json_object["test_ratio"],
            batch_size=json_object["batch_size"],
            shuffle=json_object["shuffle"],
        )

        evaluation = EVALUATION_MAP[json_object["evaluation"]](
            dataset_name=json_object["dataset_name"],
            dataset_save_path=json_object["dataset_save_path"],
            test_ratio=json_object["test_ratio"],
            model_net=json_object["model"],
            optimizer=json_object["optimizer"],
            data_set=data_set,
            result_save_dir=json_object["result_save_dir"],
            model_save_freq=json_object["model_save_freq"],
        )
        json_object["override"] = convert_str_to_bool(json_object["override"])

        return Config(
            model=MODEL_MAP[json_object["model"]],
            optimizer=json_object["optimizer"],
            Evaluation=evaluation,
            data_set=data_set,
            # model parameters
            dataset_name=str(json_object["dataset_name"]),
            num_classes=int(json_object["num_classes"]),
            num_features=int(json_object["num_features"]),
            net_dimension=int(json_object["net_dimension"]),
            learning_rate=float(json_object["learning_rate"]),
            # evaluation parameters
            dataset_save_path=str(json_object["dataset_save_path"]),
            test_ratio=int(json_object["test_ratio"]),
            batch_size=int(json_object["batch_size"]),
            shuffle=bool(json_object["shuffle"]),
            num_cv=int(json_object["num_cv"]),
            num_epoch=int(json_object["num_epoch"]),
            result_save_dir=str(json_object["result_save_dir"]),
            model_save_freq=int(json_object["model_save_freq"]),
            override=json_object["override"],
        )



"""
File for data loading from the standard datasets implemented in the pytorch_geometric #
library. The DataSet loader is implemented as a base class and other subclasses include loaders for standardized benchmarks
as well as custom datasets.

:author: Julian M. Kleber
"""
from typing import Type, Optional, List
from abc import ABC, abstractclassmethod, abstractmethod

import torch
import torch_geometric
from torch_geometric.data import DataLoader
from torch_geometric.datasets import MoleculeNet, TUDataset

import rdkit as rdkit

from carate.default_interface import DefaultObject

import logging

logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class DatasetObject(ABC, DefaultObject, torch.utils.data.Dataset):
    """
    Interface for DataLoading objects
    """

    def __init__(
        self,
        dataset_name: str,
        dataset_save_path: str,
        test_ratio: int,
        batch_size: int,
        shuffle: bool,
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractclassmethod
    def load_data(
        self,
        dataset_name: str,
        dataset_save_path: str,
        test_ratio: int,
        batch_size: int,
        shuffle: bool,
    ) -> None:
        raise NotImplementedError  # pragma: no cover

    @abstractmethod
    def __repr__(self):
        raise NotImplementedError  # pragme: no cover


class StandardPytorchGeometricDataset(DatasetObject):

    DataSet: torch.utils.data.Dataset

    @classmethod
    def load_data(
        cls,
        dataset_name: str,
        test_ratio: int,
        dataset_save_path: str,
        batch_size: int = 64,
        shuffle: bool = True,
    ) -> List[torch.utils.data.DataLoader | torch.utils.data.Dataset]:
        """
        The load_dataset function loads a standard dataset, splits it into a training and testing set,
        and returns the appropriate dataloaders for each. The test_ratio parameter specifies what percentage of
        the original dataset should be used as the testing set. The batch_size parameter specifies how many samples
        should be in each batch.

        :param path:str: Used to Define the path where the dataset is located.
        :param dataset_name:str: Used to Specify which dataset to load.
        :param test_ratio:int: Used to divide the dataset into a training and test set.
        :param batch_size:int: Used to set the batch size for training.
        :return: A train_loader and a test_loader.

        :doc-author: Julian M. Kleber
        """

        if shuffle:
            dataset = cls.DataSet(
                dataset_save_path, name=dataset_name).shuffle()
        else:
            dataset = cls.DataSet(dataset_save_path, name=dataset_name)

        test_dataset = dataset[: len(dataset) // test_ratio]
        train_dataset = dataset[len(dataset) // test_ratio:]
        test_loader = DataLoader(test_dataset, batch_size=batch_size)
        train_loader = DataLoader(train_dataset, batch_size=batch_size)

        return train_loader, test_loader, dataset, train_dataset, test_dataset


class StandardDatasetMoleculeNet(StandardPytorchGeometricDataset):
    """
    Implementation of the Dataset interaface with focus on the models implemented in pytorch_geometric
    and provided by the MoleculeNet collection of datasets.
    """

    DataSet = MoleculeNet

    def __init__(
        self,
        dataset_save_path: str,
        dataset_name: str,
        test_ratio: int,
        batch_size: int,
        shuffle: bool = True,
    ):
        """
        The __init__ function is called the constructor and is automatically called when you create a new instance of this class.
        The __init__ function allows us to set attributes that are specific to each object created from the class.
        In our case, we want each data_set object to have a path, dataset_name, test_ratio and batch size attribute.

        :param self: Used to Reference the object to which the function is applied.
        :param path:str: Used to Specify the path to the dataset.
        :param dataset_name:str: Used to Store the name of the data set.
        :param test_ratio:int: Used to Split the data set into a training and testing set.
        :param batch_size:int: Used to Set the batch size.
        :return: The object of the class.

        :doc-author: Julian M. Kleber
        """

        self.dataset_save_path = dataset_save_path
        self.dataset_name = dataset_name
        self.test_ratio = test_ratio
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __repr__(self):
        return "StandardMoleculeNet"


class StandardDatasetTUDataset(StandardPytorchGeometricDataset):
    """
    class for loading standard datasates from the TU Dataset collection implemented
    by PyTorch Geometric.

    author: Julian M. Kleber
    """

    DataSet = TUDataset

    def __init__(
        self,
        dataset_save_path: str,
        dataset_name: str,
        test_ratio: int,
        batch_size: int,
        shuffle: bool = True,
    ):
        """
        The __init__ function is called the constructor and is automatically called when you create a new instance of this class.
        The __init__ function allows us to set attributes that are specific to each object created from the class.
        In our case, we want each data_set object to have a path, dataset_name, test_ratio and batch size attribute.

        :param self: Used to Reference the object to which the function is applied.
        :param path:str: Used to Specify the path to the dataset.
        :param dataset_name:str: Used to Store the name of the data set.
        :param test_ratio:int: Used to Split the data set into a training and testing set.
        :param batch_size:int: Used to Set the batch size.
        :return: The object of the class.

        :doc-author: Julian M. Kleber
        """

        self.dataset_save_path = dataset_save_path
        self.dataset_name = dataset_name
        self.test_ratio = test_ratio
        self.batch_size = batch_size
        self.shuffle = shuffle

    def __repr__(self):
        return "StandardTUDataset"



"""
This is the heart of the application and trains / tests a algorithm on a given dataset.
The idea is to parametrize as much as possible.

:author: Julian M. Kleber
"""
from typing import Type, Any, Dict
import json
import numpy as np

import logging

from sklearn import metrics
import torch
import torch.nn.functional as F
from amarium.utils import check_make_dir, prepare_file_name_saving

from carate.utils.model_files import (
    save_model_training_checkpoint,
    save_model_parameters,
    load_model_training_checkpoint,
)

from carate.load_data import (
    DatasetObject,
)
from carate.default_interface import DefaultObject
from carate.models.base_model import Model


logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class Evaluation(DefaultObject):

    """
    The evaluation class is about evaluating a given model written in PyTorch or PyTorchGeometric.
    """

    name = "Default evaluation"

    def __init__(
        self,
        dataset_name: str,
        dataset_save_path: str,
        result_save_dir: str,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        data_set: DatasetObject,
        test_ratio: int,
        num_epoch: int = 150,
        num_cv: int = 5,
        num_classes: int = 2,
        out_dir: str = r"./out",
        batch_size: int = 64,
        shuffle: bool = True,
        model_save_freq: int = 100,
        override: bool = True,
    ) -> None:
        """

        :param self: Used to Refer to the object instance itself, and is used to access variables that belongs to the class.
        :param model: Used to Specify the model that will be trained.
        :param optimizer: Used to Define the optimizer that will be used to train the model.
        :param data_set:Type[DatasetObject]: Used to Specify the type of data loader that is used. Is implemented according to
                                             the interface given in load_data.py by the class DatasetObject.load_data().

        :param epoch:int=150: Used to Set the number of epochs to train for.
        :param num_cv:int=5: Used to Specify the number of cross validations that will be used in the training process.
        :param num_classes:int=2: Used to Define the number of classes in the dataset.
        :param out_dir:str="out": Used to Specify the directory where the output of your training will be stored.
        :return: The following:.

        :doc-author: Julian M. Kleber
        """
        self.dataset_name = dataset_name
        self.dataset_save_path = dataset_save_path
        self.test_ratio = test_ratio
        self.num_epoch = num_epoch
        self.model_net = model_net
        self.optimizer = optimizer
        self.num_classes = num_classes
        self.num_cv = num_cv
        self.out_dir = out_dir
        self.data_set = data_set
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu")
        self.result_save_dir = result_save_dir
        self.model_save_freq = model_save_freq
        self.override = override
        self.train_store = None

    def cv(
        self,
        num_cv: int,
        num_epoch: int,
        num_classes: int,
        dataset_name: str,
        dataset_save_path: str,
        test_ratio: int,
        data_set: DatasetObject,
        shuffle: bool,
        batch_size: int,
        model_net: Model,
        optimizer: torch.optim.Optimizer,
        device: torch.device,
        result_save_dir: str,
        model_save_freq: int,
        override: bool = True,
    ) -> Dict[str, Any]:
        """
        The function is the core of the evaluation. The results are saved on disk during
        the run and returned as json at the end of the run.

        :param self: Used to Represent the instance of the class.
        :param num_cv:int: Used to Specify the number of cross-validation folds.
        :param num_epoch:int: Used to Specify the number of epochs to train for.
        :param num_classes:int: Used to Determine the number of classes in the dataset.
        :param dataset_name:str: Used to Specify the name of the dataset to be used.
        :param DataSetType[DatasetObject]: Used to Load the data.
        :param : Used to Specify the number of folds in a (stratified)kfold,.
        :return: A list of dictionaries.

        :doc-author: Trelent
        """

        (
            num_cv,
            num_epoch,
            num_classes,
            dataset_name,
            dataset_save_path,
            test_ratio,
            data_set,
            shuffle,
            batch_size,
            model_net,
            optimizer,
            device,
            result_save_dir,
            model_save_freq,
            override,
        ) = self._get_defaults(locals())
        result = []

        tmp = {}
        save_model_parameters(model_net, save_dir=result_save_dir)
        for i in range(num_cv):
            (
                train_loader,
                test_loader,
                loaded_dataset,
                train_dataset,
                test_dataset,
            ) = data_set.load_data(
                dataset_name=dataset_name,
                dataset_save_path=dataset_save_path,
                test_ratio=test_ratio,
                batch_size=batch_size,
                shuffle=shuffle,
            )
            # storage containers

            acc_store_train = []
            acc_store_test = []
            auc_store = []
            loss_store = []

            for epoch in range(1, num_epoch + 1):

                train_loss = self.train(
                    epoch=epoch,
                    model_net=model_net,
                    device=device,
                    optimizer=optimizer,
                    train_loader=train_loader,
                    num_classes=num_classes,
                )
                loss_store.append(train_loss.cpu().tolist())
                train_acc = self.test(
                    train_loader,
                    device=device,
                    model_net=model_net,
                    epoch=epoch,
                    test=False,
                )  # test False for storing the results
                test_acc, self.train_store = self.test(
                    test_loader,
                    device=device,
                    model_net=model_net,
                    epoch=epoch,
                    test=True,
                )
                acc_store_train.append(train_acc.cpu().tolist())
                acc_store_test.append(test_acc.cpu().tolist())
                logging.info(
                    "Epoch: {:03d}, Train Loss: {:.7f}, Train Acc: {:.7f}, Test Acc: {:.7f}".format(
                        epoch, train_loss, train_acc, test_acc
                    )
                )
                y = np.zeros((len(test_dataset)))
                x = self.train_store

                for j in range(len(test_dataset)):
                    y[j] = test_dataset[j].y

                y = torch.as_tensor(y)
                y = F.one_hot(y.long(), num_classes=num_classes)
                store_auc = []

                for j in range(len(x[0, :])):
                    auc = metrics.roc_auc_score(y[:, j], x[:, j])
                    logging.info("AUC of " + str(j) + "is:" + str(auc))
                    store_auc.append(auc)

                auc_store.append(store_auc)

                tmp["Loss"] = list(loss_store)
                tmp["Acc_train"] = list(acc_store_train)
                tmp["Acc_test"] = list(acc_store_test)
                tmp["AUC"] = list(auc_store)

                if epoch % model_save_freq == 0:

                    self.save_whole_checkpoint(
                        result_save_dir=result_save_dir,
                        dataset_name=dataset_name,
                        num_cv=i,
                        num_epoch=epoch,
                        model_net=model_net,
                        data=tmp,
                        optimizer=optimizer,
                        loss=train_loss,
                        override=override,
                    )
            result.append(tmp)
        return result

    def train(
        self,
        epoch: int,
        model_net: Model,
        device: torch.device,
        train_loader: Type[torch.utils.data.DataLoader],
        optimizer: torch.optim.Optimizer,
        num_classes: int,
    ):
        """
        The train function is used to train the model.
           The function takes in a number of epochs and a model, and returns the accuracy on the test set.

        :param epoch: Used to Determine when to stop training.
        :param model: Used to Pass the model to the function.
        :param device: Used to Tell the model which device to use.
        :param train_loader: Used to Load the training data.
        :param test_loader: Used to Evaluate the model on the test data.
        :param optimizer: Used to Specify the optimizer that will be used in training.
        :param num_classes=2: Used to Specify the number of classes in the data.
        :param shrikage=51: Used to Make sure that the model is trained for at least 51 epochs.
        :return: The accuracy of the model on the training set.

        :doc-author: Trelent
        """
        model_net.train()

        correct = 0
        for data in train_loader:
            data.x = data.x.type(torch.FloatTensor)
            data.y = F.one_hot(data.y.long(), num_classes=num_classes).type(
                torch.FloatTensor
            )
            data = data.to(device)
            optimizer.zero_grad()
            output_probs = model_net(data.x, data.edge_index, data.batch)
            output = (output_probs > 0.5).float()
            loss = torch.nn.BCELoss()
            loss = loss(output_probs, data.y)
            loss.backward()
            optimizer.step()
            correct += (output == data.y).float().sum() / num_classes
        accuracy = correct / len(train_loader.dataset)
        return accuracy

    def test(
        self,
        test_loader: torch.utils.data.DataLoader,
        epoch: int,
        model_net: Model,
        device: torch.device,
        **kwargs: Any,
    ) -> Any:
        """
        The test function is used to test the model on a dataset.
        It returns the accuracy of the model on that dataset calculated as
        the average of the atomic accuracy for each batch in the Dataset

        :param test_loader: Used to pass the test data loader.
        :param epoch: Used to keep track of the current epoch.
        :param model_net: Used to pass the model to the test function.
        :param device: Used to tell torch which device to use.
        :param test=False: Used to distinguish between training
        and testing.
        :return: The accuracy of the model on the test data.

        :doc-author: Julian M. Kleber
        """
        test = bool(kwargs["test"])
        model_net.eval()
        correct = 0
        if test:
            outs = []
        for data in test_loader:
            data.x = data.x.type(torch.FloatTensor)
            data = data.to(device)
            output_probs = model_net(data.x, data.edge_index, data.batch)
            output = (output_probs > 0.5).float()
            correct += (torch.argmax(output, dim=1) == data.y).float().sum()
            if test:
                outs.append(output.cpu().detach().numpy())
        if test:
            outputs = np.concatenate(outs, axis=0).astype(float)
            return (
                correct / len(test_loader.dataset),
                outputs,
            )  # TODO this is from some "quick experiment" not sure if the line and functionality is
        return correct / len(test_loader.dataset)

    def save_result(
        self,
        result_save_dir: str,
        dataset_name: str,
        num_cv: int,
        num_epoch: int,
        data: dict,
    ) -> None:
        """
        The save_result function saves the results of a cross-validation run to a .json file. The goal is to provide
        a json interface of cv results for later analysis of the training runs.


        :param self: Used to represent the instance of the class.
        :param result_save_dir:str: Used to specify the directory where the results will be saved.
        :param dataset_name:str: Used to identify the dataset.
        :param num_cv:int: Used to specify the number of cross validation runs.
        :param num_epoch int: Epoch the run was saved in
        :param data:dict: Used to store the results of each cross validation run.
        :return: None.

        :doc-author: Julian M. Kleber
        """

        prefix = result_save_dir + "/data/" + "CV_" + str(num_cv)
        file_name = prepare_file_name_saving(
            prefix=prefix,
            file_name=dataset_name,
            suffix=".json",
        )
        with open(file_name, "w") as f:
            json.dump(data, f)
            logging.info(
                "Saved"
                + str(num_epoch)
                + "of cv"
                + str(num_cv)
                + " run to "
                + result_save_dir
                + dataset_name
                + "_"
                + str(num_cv)
                + ".csv"
            )

    def save_whole_checkpoint(
        self,
        result_save_dir: str,
        dataset_name: str,
        num_cv: int,
        num_epoch: int,
        model_net: Type[torch.nn.Module],
        data: dict,
        optimizer: Type[torch.optim.Optimizer],
        loss: float,
        override: bool = True,
    ) -> None:
        """
        The save_whole_checkpoint function saves the model checkpoint and results for a given epoch.

        The save_whole_checkpoint function saves the model checkpoint and results for a given epoch. It is called by the train function in order to save checkpoints at regular intervals during training, as well as after each cross-validation fold has been trained on. The saved files are used to resume training if it is interrupted, or to evaluate performance of different models on test data without having to retrain them from scratch.

        :param self: Used to Represent the instance of the class.
        :param result_save_dir:str: Used to Specify the directory where the checkpoint will be saved.
        :param dataset_name:str: Used to Name the dataset.
        :param num_cv:int: Used to Specify the cross validation number.
        :param num_epoch:int: Used to Specify the number of epochs that have been completed.
        :param model_net:Type[torch.nn.Module]: Used to Save the model.
        :param data:dict: Used to Save the data, which is a dictionary containing the training and validation data.
        :param optimizer:Type[torch.optim.Optimizer]: Used to Save the optimizer state.
        :param loss:float: Used to Save the loss value.
        :param override:bool=True: Used to Override the previous checkpoint.
        :param : Used to Save the model.
        :return: None.

        :doc-author: Julian M. Kleber
        """

        self.save_model_checkpoint(
            result_save_dir=result_save_dir,
            dataset_name=dataset_name,
            num_cv=num_cv,
            num_epoch=num_epoch,
            model_net=model_net,
            optimizer=optimizer,
            loss=loss,
            override=override,
        )

        self.save_result(
            result_save_dir=result_save_dir,
            dataset_name=dataset_name,
            data=data,
            num_cv=num_cv,
            num_epoch=num_epoch,
        )
        logging.info(
            f"Successfully saved a checkpoint for epoch {num_epoch} in CV {num_cv}"
        )

    def save_model_checkpoint(
        self,
        result_save_dir: str,
        dataset_name: str,
        num_cv: int,
        num_epoch: int,
        model_net: Type[torch.nn.Module],
        optimizer: Type[torch.optim.Optimizer],
        loss: float,
        override: bool = True,
    ) -> None:
        """
        The save_model function saves the model to a file.

        The save_model function saves the model to a file. The filename is based on
        the dataset name, number of cross-validation folds, and epoch number. The
        file is saved in the result_save_dir directory with an extension of .pt (for
        PyTorch). If this directory does not exist, it will be created before saving
        the file.

        :param result_save_dir:str: Used to specify the directory where the model will be saved.
        :param dataset_name:str: Used to save the model with a name that includes the dataset it was trained on.
        :param num_cv:int: Used to specify which cross validation fold the model is being saved for.
        :param num_epoch:int: Used to save the model at a certain epoch.
        :param model_net:Type[torch.nn.Module]: Used to save the model.
        :param : Used to save the model at a certain frequency.
        :return: None.

        :doc-author: Julian M. Kleber
        """

        save_model_training_checkpoint(
            result_save_dir=result_save_dir,
            dataset_name=dataset_name,
            num_cv=num_cv,
            num_epoch=num_epoch,
            model_net=model_net,
            optimizer=optimizer,
            loss=loss,
            override=override,
        )

    def load_model_checkpoint(
        self,
        checkpoint_path: str,
        model_net: Model,
        optimizer=torch.optim.Optimizer,
    ) -> Model:
        """
        The load_model_checkpoint function loads a model checkpoint from the specified path.

        The function loads a model checkpoint from the specified path, and sets it as the
        model of this evaluation object. The function also returns that loaded model.

        :param self: Used to Refer to the object itself.
        :param checkpoint_path:str: Used to Specify the path to the checkpoint file.
        :param model_net:Model: Used to Specify the model that is being loaded.
        :param optimizer=torch.optim.Optimizer: Used to Load the optimizer state
        from a checkpoint.
        :param : Used to Load the model checkpoint.
        :return: The model.

        :doc-author: Julian M. Kleber
        """

        model_net_cp = load_model_training_checkpoint(
            checkpoint_path=checkpoint_path, model_net=model_net, optimizer=optimizer
        )
        self.model_net = (
            model_net_cp  # set the model of the evaluation object to the checkpoint
        )
        return model_net_cp

    def __str__(self):
        return "Evaluation for " + str(self.model_net) + " with the " + self.name

    def __repr__(self):
        return "Standard Evaluation Object"



"""
cgc_regression model is named after the structure of the graph neural network.
The graph neural network is structured with a convolutional , graph attention,
and another convolutional layer. The cgc_classificatin model was the model tested int the publication
Introducing CARATE: Finally speaking chemistry.
"""
import numpy as np
import torch
import torch.nn.functional as F
from torch.nn import Linear
from torch_geometric.nn import global_add_pool, GraphConv, GATConv
import sklearn.metrics as metrics

import logging

from carate.models.base_model import Model


logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class Net(Model):
    def __init__(self, dim: int, num_classes: int, num_features: int) -> None:

        super(Net, self).__init__(
            dim=dim, num_classes=num_classes, num_features=num_features
        )

        self.conv1 = GraphConv(self.num_features, self.dim)
        self.conv3 = GATConv(self.dim, self.dim, dropout=0.6)
        self.conv5 = GraphConv(self.dim, self.dim)

        self.fc1 = Linear(self.dim, self.dim)
        self.fc2 = Linear(self.dim, self.num_classes)

    def forward(self, x: float, edge_index: int, batch: int, edge_weight=None) -> float:
        x = F.relu(self.conv1(x, edge_index, edge_weight))
        x = F.dropout(x, p=0.5, training=self.training)
        x = F.relu(self.conv3(x, edge_index, edge_weight))
        x = F.relu(self.conv5(x, edge_index, edge_weight))
        x = global_add_pool(x, batch)
        x = F.relu(self.fc1(x))
        x = F.dropout(x, p=0.5, training=self.training)
        x = self.fc2(x)
        return x



"""
Module for parsing training results

:author: Julian M. Kleber
"""
from typing import Type, Dict, Any, List

from amarium.utils import load_json_from_file


def load_training_from_json_file(file_path: str) -> Dict[Any, Any]:
    """
    The load_training_from_json_file function loads a training result from a JSON file.


    :param file_path:str: Used to Specify the path to the file that you want to load.
    :return: A dictionary.

    :doc-author: Trelent
    """

    training_result: Dict[Any, Any] = load_json_from_file(file_path=file_path)
    return training_result


def get_loss_json(json_object: Dict[Any, List[float]]) -> List[float]:
    """
    The get_loss_json function takes in a json object and returns the loss value from that json object.

    :param json_object:dict: Used to Specify that the function takes a dictionary as an argument.
    :return: A list of dictionaries.

    :doc-author: Trelent
    """

    loss = json_object["Loss"]
    return loss


def get_accuracy(json_object: Dict[Any, List[float]]) -> List[float]:
    """
    The get_accuracy function takes in a json object and returns the accuracy of the model.
        Args:
            json_object (dict): A dictionary containing all of the information from a single run.

        Returns:
            List[float]: The accuracy for each epoch during training.

    :param json_object:dict: Used to Specify the type of the parameter.
    :return: A list of floats.

    :doc-author: Trelent
    """

    acc = json_object["Acc"]
    return acc


def get_auc(json_object: Dict[Any, List[float]]) -> List[float]:

    auc = json_object["AUC"]
    return auc



"""
cgc_classification model is named after the structure of the graph neural network.
The graph neural network is structured with a convolutional , graph attention,
and another convolutional layer. The cgc_classificatin model was the model 
tested int the publication Introducing CARATE: Finally speaking chemistry.

:author: Julian M. Kleber
"""
import torch
import torch.nn.functional as F
from torch.nn import Linear

from torch_geometric.nn import global_add_pool, GraphConv, GATConv

import numpy as np
import sklearn.metrics as metrics

import logging


from carate.models.base_model import Model

logger = logging.getLogger(__name__)
logging.basicConfig(
    filename="train.log",
    encoding="utf-8",
    level=logging.DEBUG,
    format="%(asctime)s %(message)s",
)


class Net(Model):
    """
    The Net is the core algorithm and needs a constructor and a
    forward pass. The train, test and evaluation methods are implemented
    in the evaluation module with the Evaluation class.

    """

    def __init__(self, dim: int, num_features: int, num_classes: int) -> None:
        super(Net, self).__init__(
            dim=dim, num_features=num_features, num_classes=num_classes
        )

        self.conv1 = GraphConv(self.num_features, self.dim)
        self.conv3 = GATConv(self.dim, self.dim, dropout=0.6, heads=16)
        self.conv5 = GraphConv(self.dim * 16, self.dim)

        self.fc1 = Linear(self.dim, self.dim)
        self.fc2 = Linear(self.dim, self.num_classes)

    def forward(self, x, edge_index, batch, edge_weight=None):
        x = F.relu(self.conv1(x, edge_index, edge_weight))
        x = F.dropout(x, p=0.5, training=self.training)
        x = F.relu(self.conv3(x, edge_index, edge_weight))
        x = F.relu(self.conv5(x, edge_index, edge_weight))
        x = global_add_pool(x, batch)
        x = F.relu(self.fc1(x))
        x = F.dropout(
            x, p=0.5, training=self.training
        )  # TODO thats not a good dropout and should be reduced
        x = self.fc2(x)
        return torch.sigmoid(x)



"""
Utility file for model checkpoints file operations

:author: Julian M. Kleber
"""
import os

import torch

from amarium.utils import (
    prepare_file_name_saving,
    load_json_from_file,
    make_full_filename,
    save_json_to_file,
    check_make_dir,
)

from typing import Tuple, Type, Dict, Any

from carate.models.base_model import Model


def load_model(
    model_path: str, model_net: Type[torch.nn.Module]
) -> Type[torch.nn.Module]:
    """
    The load_model function takes in a model_path, model_params_path and the type of network to be loaded.
    It then loads the parameters from the params file into a dictionary and uses that to create an instance of
    the specified network. It then loads in the state dict from PATH and sets it as eval mode.

    :param model_path:str: Used to specify the path to the model file.
    :param model_params_path:str: Used to load the model parameters from a file.
    :param model_net:Type[torch.nn.Module]: Used to specify the type of model that is being loaded.
    :return: A model that is loaded with the parameters in the path.

    :doc-author: Julian M. Kleber
    """

    model = model_net()
    model.load_state_dict(torch.load(model_path))
    model.eval()
    return model


def load_model_training_checkpoint(
    checkpoint_path: str,
    model_net: Type[torch.nn.Module],
    optimizer: Type[torch.optim.Optimizer],
) -> Tuple[Model, torch.optim.Optimizer]:

    # For any bug fixing please consult the PyTorch documentation:  https://pytorch.org/tutorials/beginner/saving_loading_models.html#saving-loading-a-general-checkpoint-for-inference-and-or-resuming-training

    model = model_net
    optimizer = optimizer
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optimizer.load_state_dict(checkpoint["optimizer_state_dict"])
    epoch = checkpoint["epoch"]
    loss = checkpoint["loss"]
    model.train()
    return model, optimizer


def save_model_training_checkpoint(
    result_save_dir: str,
    dataset_name: str,
    num_cv: int,
    num_epoch: int,
    model_net: Type[torch.nn.Module],
    optimizer: Type[torch.optim.Optimizer],
    loss: float,
    override: bool,
) -> None:

    # For any bug fixing please refer to https://pytorch.org/tutorials/beginner/saving_loading_models.html#saving-loading-a-general-checkpoint-for-inference-and-or-resuming-training

    prefix = result_save_dir + "/checkpoints/CV_" + str(num_cv)
    check_make_dir(prefix)

    if override is True:
        save_path = prepare_file_name_saving(
            prefix=prefix,
            file_name=dataset_name,
            suffix=".tar",
        )
    if override is False:
        save_path = prepare_file_name_saving(
            prefix=prefix,
            file_name=dataset_name + "_Epoch-" + str(num_epoch),
            suffix=".tar",
        )
    torch.save(
        {
            "epoch": num_epoch,
            "cv": num_cv,
            "model_state_dict": model_net.state_dict(),
            "optimizer_state_dict": optimizer.state_dict(),
            "loss": loss,
        },
        save_path,
    )


def save_model(
    result_save_dir: str,
    dataset_name: str,
    num_cv: int,
    num_epoch: int,
    model_net: Type[torch.nn.Module],
) -> None:
    """
    The save_model function saves the model to a file.

    The save_model function saves the model to a file. The filename is constructed from the dataset name, number of cross-validation folds, and number of epochs trained on.

    :param result_save_dir:str: Used to specify the directory where the model will be saved.
    :param dataset_name:str: Used to name the file.
    :param num_cv:int: Used to make the filename unique.
    :param num_epoch: Used to save the model after a certain number of epochs.
    :param model_net:Type[torch.nn.Module]: Used to save the model.
    :param : Used to specify the directory where the model will be saved.
    :return: The path of the saved model.

    :doc-author: Julian M. Kleber
    """
    prefix = result_save_dir + "/checkpoints/CV_" + str(num_cv)
    check_make_dir(prefix)
    save_path = prepare_file_name_saving(
        prefix=prefix,
        file_name=dataset_name + "_Epoch-" + str(num_epoch),
        suffix=".pt",
    )
    torch.save(model_net.state_dict(), save_path)


def load_model_parameters(model_params_file_path: str) -> Dict[Any, Any]:
    """
    The load_model_parameters function loads the model parameters from a JSON file.

    Parameters:

        model_params_file_path (str): The path to the JSON file containing the model parameters.

    Returns:

        dict: A dictionary of all of the loaded model parameters.

    :param model_params_file_path:str: Used to Specify the file path of the model parameters.
    :return: A dictionary of model parameters.

    :doc-author: Julian M. Kleber
    """

    return load_json_from_file(model_params_file_path)


def save_model_parameters(model_net: Model, save_dir: str) -> None:
    """
    The save_model_parameters function saves the model architecture to a csv file.

        Args:
            model_net (torch.nn.Module): The neural network that is being used for training and testing, e.g., CNN() or RNN().
            save_path (str): The path where the json file will be saved to, e.g., "./model/".

        Returns: None

    :param model_net:Type[torch.nn.Module]: Used to specify the type of model that is being used.
    :param save_path:str: Used to save the model architecture in a json file.
    :return: A dictionary of the model architecture (model_architecture).

    :doc-author: Julian M. Kleber
    """
    prefix = save_dir + "/model_parameters/"
    model_architecture = model_net.__dict__
    file_name = prepare_file_name_saving(
        prefix=prefix, file_name="model_architecture", suffix=".json"
    )
    save_json_to_file(model_architecture, file_name=file_name)


def get_latest_checkpoint(search_dir: str, num_cv: int, epoch: int) -> str:

    if not search_dir.endswith("/"):
        search_dir += "/"

    search_dir += "checkpoints"
    checkpoint_dirs = os.listdir(search_dir)
    correct_sub_dir = checkpoint_dirs[checkpoint_dirs.index(
        "CV_" + str(num_cv))]
    search_dir += "/" + correct_sub_dir
    checkpoints = os.listdir(search_dir)
    checkpoints = sorted(checkpoints)
    return checkpoints[:-1]
